import pytest
from pytest import approx
import pymap3d as pm

lat, lon = (65, -148)
lla0 = (42, -82, 200)
azel = (180.1, 80)
t0 = '2014-04-06T08:00:00'
radec = (166.5032081149338, 55.000011165405752)


@pytest.mark.parametrize('usevallado', [True, False])
def test_azel2radec(usevallado):
    radec1 = pm.azel2radec(*azel, lat, lon, t0, usevallado=usevallado)
    assert radec1 == approx(radec, rel=0.01)


@pytest.mark.parametrize('usevallado', [True, False])
def test_radec2azel(usevallado):
    azel1 = pm.radec2azel(*radec, lat, lon, t0, usevallado=usevallado)
    assert azel1 == approx(azel, rel=0.01)


if __name__ == '__main__':
    pytest.main(['-xrsv', __file__])



"""
Azimuth / elevation <==> Right ascension, declination
"""
from typing import Tuple
from datetime import datetime
from .vallado import azel2radec as vazel2radec, radec2azel as vradec2azel
from .timeconv import str2dt  # astropy can't handle xarray times (yet)
try:
    from astropy.time import Time
    from astropy import units as u
    from astropy.coordinates import Angle, SkyCoord, EarthLocation, AltAz, ICRS
except ImportError:
    Time = None


def azel2radec(az_deg: float, el_deg: float,
               lat_deg: float, lon_deg: float,
               time: datetime, usevallado: bool = False) -> Tuple[float, float]:
    """
    viewing angle (az, el) to sky coordinates (ra, dec)

    Parameters
    ----------
    az_deg : float or numpy.ndarray of float
         azimuth [degrees clockwize from North]
    el_deg : float or numpy.ndarray of float
             elevation [degrees above horizon (neglecting aberration)]
    lat_deg : float
              observer latitude [-90, 90]
    lon_deg : float
              observer longitude [-180, 180] (degrees)
    time : datetime.datetime or str
           time of observation
    usevallado : bool, optional
                 default use astropy. If true, use Vallado algorithm

    Returns
    -------
    ra_deg : float or numpy.ndarray of float
         ecliptic right ascension (degress)
    dec_deg : float or numpy.ndarray of float
         ecliptic declination (degrees)
    """

    if usevallado or Time is None:  # non-AstroPy method, less accurate
        return vazel2radec(az_deg, el_deg, lat_deg, lon_deg, time)

    obs = EarthLocation(lat=lat_deg * u.deg, lon=lon_deg * u.deg)

    direc = AltAz(location=obs, obstime=Time(str2dt(time)),
                  az=az_deg * u.deg, alt=el_deg * u.deg)

    sky = SkyCoord(direc.transform_to(ICRS()))

    return sky.ra.deg, sky.dec.deg


def radec2azel(ra_deg: float, dec_deg: float,
               lat_deg: float, lon_deg: float,
               time: datetime, usevallado: bool = False) -> Tuple[float, float]:
    """
    sky coordinates (ra, dec) to viewing angle (az, el)

    Parameters
    ----------
    ra_deg : float or numpy.ndarray of float
         ecliptic right ascension (degress)
    dec_deg : float or numpy.ndarray of float
         ecliptic declination (degrees)
    lat_deg : float
              observer latitude [-90, 90]
    lon_deg : float
              observer longitude [-180, 180] (degrees)
    time : datetime.datetime or str
           time of observation
    usevallado : bool, optional
                 default use astropy. If true, use Vallado algorithm

    Returns
    -------
    az_deg : float or numpy.ndarray of float
             azimuth [degrees clockwize from North]
    el_deg : float or numpy.ndarray of float
             elevation [degrees above horizon (neglecting aberration)]
    """

    if usevallado or Time is None:
        return vradec2azel(ra_deg, dec_deg, lat_deg, lon_deg, time)

    obs = EarthLocation(lat=lat_deg * u.deg,
                        lon=lon_deg * u.deg)

    points = SkyCoord(Angle(ra_deg, unit=u.deg),
                      Angle(dec_deg, unit=u.deg),
                      equinox='J2000.0')

    altaz = points.transform_to(AltAz(location=obs, obstime=Time(str2dt(time))))

    return altaz.az.degree, altaz.alt.degree



import pytest
from pytest import approx
import pymap3d as pm

t0 = '2014-04-06T08:00:00'
t1 = '2013-01-15T12:00:05'
eci0 = (-3.977913815668146e6, -2.582332196263046e6, 4.250818828152067e6)


@pytest.mark.parametrize('useastropy', [True, False])
def test_eciecef(useastropy):
    ecef = pm.eci2ecef(*eci0, t1, useastropy=useastropy)
    assert ecef == approx([649012.04640917, -4697980.55129606, 4250818.82815207], rel=0.001)

    assert pm.ecef2eci(*ecef, t1, useastropy=useastropy) == approx(eci0, rel=0.001)


@pytest.mark.parametrize('useastropy', [True, False])
def test_eci_times(useastropy):
    with pytest.raises(ValueError):
        pm.eci2ecef(*eci0, [t0, t0], useastropy=useastropy)

    with pytest.raises(ValueError):
        pm.ecef2eci(*eci0, [t0, t0], useastropy=useastropy)

    x = [eci0[0]] * 2
    y = [eci0[1]] * 2
    z = [eci0[2]] * 2
    t = [t0] * 2
    assert pm.ecef2eci(*pm.eci2ecef(x, y, z, t, useastropy=useastropy),
                       t, useastropy=useastropy) == approx(eci0, rel=0.001)


if __name__ == '__main__':
    pytest.main(['-x', __file__])



#!/usr/bin/env python
import pytest
from pytest import approx
import pymap3d as pm

xyz0 = (660e3, -4700e3, 4247e3)


def test_ellipsoid():

    assert pm.ecef2geodetic(*xyz0, ell=pm.Ellipsoid('wgs84')) == approx([42.014670535, -82.0064785, 276.9136916])
    assert pm.ecef2geodetic(*xyz0, ell=pm.Ellipsoid('grs80')) == approx([42.014670536, -82.0064785, 276.9137385])
    assert pm.ecef2geodetic(*xyz0, ell=pm.Ellipsoid('clarke1866')) == approx([42.01680003, -82.0064785, 313.9026793])
    assert pm.ecef2geodetic(*xyz0, ell=pm.Ellipsoid('mars')) == approx([42.009428417, -82.006479, 2.981246e6])
    assert pm.ecef2geodetic(*xyz0, ell=pm.Ellipsoid('venus')) == approx([41.8233663, -82.0064785, 3.17878159e5])
    assert pm.ecef2geodetic(*xyz0, ell=pm.Ellipsoid('moon')) == approx([41.8233663, -82.0064785, 4.630878e6])


if __name__ == '__main__':
    pytest.main([__file__])



""" isometric latitude, meridian distance """
import numpy as np
from .ellipsoid import Ellipsoid


def isometric(lat: float, ell: Ellipsoid = None, deg: bool = True):
    """
     computes isometric latitude of a point on an ellipsoid

     Parameters
     ----------

     lat : float or numpy.ndarray of float
         geodetic latitude
     ell : Ellipsoid, optional
         reference ellipsoid (default WGS84)
     deg : bool, optional
         degrees input/output  (False: radians in/out)

     Returns
     -------

     isolat : float or numpy.ndarray of float
         isometric latiude

     Notes
     -----

     Isometric latitude is an auxiliary latitude proportional to the spacing
     of parallels of latitude on an ellipsoidal mercator projection.

     Based on Deakin, R.E., 2010, 'The Loxodrome on an Ellipsoid', Lecture Notes,
     School of Mathematical and Geospatial Sciences, RMIT University,
     January 2010
    """

    if ell is None:
        ell = Ellipsoid()

    f = ell.flattening  # flattening of ellipsoid

    if deg is True:
        lat = np.deg2rad(lat)

    e2 = f * (2 - f)  # eccentricity-squared
    e = np.sqrt(e2)  # eccentricity of ellipsoid

    x = e * np.sin(lat)
    y = (1 - x) / (1 + x)
    z = np.pi / 4 + lat / 2

#   calculate the isometric latitude
    isolat = np.log(np.tan(z) * (y**(e / 2)))

    if deg is True:
        isolat = np.degrees(isolat)

    return isolat


def meridian_dist(lat: float, ell: Ellipsoid = None, deg: bool = True):
    """
    computes the meridian distance on an ellipsoid *from the equator* to a latitude.

    Parameters
    ----------

    lat : float or numpy.ndarray of float
        geodetic latitude
    ell : Ellipsoid, optional
         reference ellipsoid (default WGS84)
    deg : bool, optional
         degrees input/output  (False: radians in/out)

    Results
    -------

    mdist : float or numpy.ndarray of float
         meridian distance (degrees/radians)


    Notes
    -----

    Formula given Baeschlin, C.F., 1948,
    "Lehrbuch Der Geodasie", Orell Fussli Verlag, Zurich, pp.47-50.

    Based on Deakin, R.E., 2010, 'The Loxodrome on an Ellipsoid', Lecture Notes,
    School of Mathematical and Geospatial Sciences, RMIT University, January 2010
    """

    if deg is True:
        lat = np.radians(lat)

    #   set ellipsoid parameters
    if ell is None:
        ell = Ellipsoid()

    a = ell.semimajor_axis
    f = ell.flattening  # flattening of ellipsoid

    e2 = f * (2 - f)  # eccentricity-squared

    # powers of eccentricity
    e4 = e2 * e2
    e6 = e4 * e2
    e8 = e6 * e2
    e10 = e8 * e2

    # coefficients of series expansion for meridian distance
    A = 1 + (3 / 4) * e2 + (45 / 64) * e4 + (175 / 256) * e6 + (11025 / 16384) * e8 + (43659 / 65536) * e10
    B = (3 / 4) * e2 + (15 / 16) * e4 + (525 / 512) * e6 + (2205 / 2048) * e8 + (72765 / 65536) * e10
    C = (15 / 64) * e4 + (105 / 256) * e6 + (2205 / 4096) * e8 + (10395 / 16384) * e10
    D = (35 / 512) * e6 + (315 / 2048) * e8 + (31185 / 131072) * e10
    E = (315 / 16384) * e8 + (3465 / 65536) * e10
    F = (693 / 131072) * e10

    term1 = A * lat
    term2 = (B / 2) * np.sin(2 * lat)
    term3 = (C / 4) * np.sin(4 * lat)
    term4 = (D / 6) * np.sin(6 * lat)
    term5 = (E / 8) * np.sin(8 * lat)
    term6 = (F / 10) * np.sin(10 * lat)

    mdist = a * (1 - e2) * (term1 - term2 + term3 - term4 + term5 - term6)

    return mdist


def loxodrome_inverse(lat1: float, lon1: float, lat2: float, lon2: float,
                      ell: Ellipsoid = None, deg: bool = True):
    """
    computes the arc length and azimuth of the loxodrome
    between two points on the surface of the reference ellipsoid

    Parameters
    ----------

    lat1 : float or numpy.ndarray of float
        geodetic latitude of first point
    lon1 : float or numpy.ndarray of float
        geodetic longitude of first point
    lat2 : float or numpy.ndarray of float
        geodetic latitude of second point
    lon2 : float or numpy.ndarray of float
        geodetic longitude of second point
    ell : Ellipsoid, optional
         reference ellipsoid (default WGS84)
    deg : bool, optional
         degrees input/output  (False: radians in/out)

    Results
    -------

    lox_s : float or numpy.ndarray of float
        distance along loxodrome
    az12 : float or numpy.ndarray of float
        azimuth of loxodrome (degrees/radians)

    Based on Deakin, R.E., 2010, 'The Loxodrome on an Ellipsoid', Lecture Notes,
    School of Mathematical and Geospatial Sciences, RMIT University, January 2010

    [1] Bowring, B.R., 1985, 'The geometry of the loxodrome on the
    ellipsoid', The Canadian Surveyor, Vol. 39, No. 3, Autumn 1985,
    pp.223-230.
    [2] Snyder, J.P., 1987, Map Projections-A Working Manual. U.S.
    Geological Survey Professional Paper 1395. Washington, DC: U.S.
    Government Printing Office, pp.15-16 and pp. 44-45.
    [3] Thomas, P.D., 1952, Conformal Projections in Geodesy and
    Cartography, Special Publication No. 251, Coast and Geodetic
    Survey, U.S. Department of Commerce, Washington, DC: U.S.
    Government Printing Office, p. 66.
    """

    #   set ellipsoid parameters
    if ell is None:
        ell = Ellipsoid()

    if deg is True:
        lat1, lon1, lat2, lon2 = np.radians([lat1, lon1, lat2, lon2])

    # compute isometric latitude of P1 and P2
    isolat1 = isometric(lat1, deg=False, ell=ell)
    isolat2 = isometric(lat2, deg=False, ell=ell)

    # compute changes in isometric latitude and longitude between points
    disolat = isolat2 - isolat1
    dlon = lon2 - lon1

    # compute azimuth
    az12 = np.arctan2(dlon, disolat)

    # compute distance along loxodromic curve
    m1 = meridian_dist(lat1, deg=False, ell=ell)
    m2 = meridian_dist(lat2, deg=False, ell=ell)
    dm = m2 - m1
    lox_s = dm / np.cos(az12)

    if deg is True:
        az12 = np.degrees(az12) % 360.

    return lox_s, az12



"""
for strictly pure Python (no Numpy) only modules
Python >= 2.7
"""
import pytest
from pytest import approx
from math import radians

import pymap3d.math as pm

lla0 = (42, -82, 200)
rlla0 = (radians(lla0[0]), radians(lla0[1]), lla0[2])
aer0 = (33, 70, 1000)
raer0 = (radians(aer0[0]), radians(aer0[1]), aer0[2])
enu0 = (186.277521, 286.842228, 939.692621)
ned0 = (enu0[1], enu0[0], -enu0[2])
vx, vy, vz = (5, 3, 2)
ve, vn, vu = (5.368859646588048, 3.008520763668120, -0.352347711524077)


def test_aer_enu():
    xyz = pm.aer2ecef(aer0[0], aer0[1], aer0[2], lla0[0], lla0[1], lla0[2])

    enu = pm.aer2enu(aer0[0], aer0[1], aer0[2])

    assert pm.enu2ecef(enu[0], enu[1], enu[2], lla0[0], lla0[1], lla0[2]) == approx(xyz)


if __name__ == '__main__':
    pytest.main([__file__])



#!/usr/bin/env python
import pytest
from pytest import approx
from math import radians, nan
import numpy as np

import pymap3d as pm

lla0 = (42, -82, 200)
rlla0 = (radians(lla0[0]), radians(lla0[1]), lla0[2])
lla1 = (42.002582, -81.997752, 1.1397018e3)
rlla1 = (np.radians(lla1[0]), np.radians(lla1[1]), lla1[2])

xyz0 = (660675.2518247,
        -4700948.68316,
        4245737.66222)

aer0 = (33, 70, 1000)
raer0 = (np.radians(aer0[0]), np.radians(aer0[1]), aer0[2])

ELL = pm.Ellipsoid()
A = ELL.semimajor_axis
B = ELL.semiminor_axis

atol_dist = 1e-6  # 1 micrometer


@pytest.mark.parametrize('lla',
                         [(42, -82, 200),
                          ([42], [-82], [200]),
                          (np.array(42), np.array(-82), np.array(200)),
                          (np.array([42]), np.array([-82]), np.array([200])),
                          (np.atleast_3d(42), np.atleast_3d(-82), np.atleast_3d(200))],
                         ids=('scalar', 'list', '0d', '1d', '3d'))
def test_scalar_geodetic2ecef(lla):
    """
    verify we can handle the wide variety of input data type users might use
    """
    x0, y0, z0 = pm.geodetic2ecef(*lla)

    assert (x0, y0, z0) == approx(xyz0)


@pytest.mark.parametrize('xyz',
                         [(xyz0[0], xyz0[1], xyz0[2]),
                          ([xyz0[0]], [xyz0[1]], [xyz0[2]]),
                          (np.array(xyz0[0]), np.array(xyz0[1]), np.array(xyz0[2])),
                          (np.array([xyz0[0]]), np.array([xyz0[1]]), np.array([xyz0[2]])),
                          (np.atleast_3d(xyz0[0]), np.atleast_3d(xyz0[1]), np.atleast_3d(xyz0[2]))],
                         ids=('scalar', 'list', '0d', '1d', '3d'))
def test_scalar_ecef2geodetic(xyz):
    """
    verify we can handle the wide variety of input data type users might use
    """
    lat, lon, alt = pm.ecef2geodetic(*xyz)

    assert [lat, lon, alt] == approx(lla0, rel=1e-4)


@pytest.mark.parametrize('xyz',
                         [(0, A, 50),
                          ([0], [A], [50]),
                          (np.array(0), np.array(A), np.array(50)),
                          (np.array([0]), np.array([A]), np.array([50])),
                          (np.atleast_3d(0), np.atleast_3d(A), np.atleast_3d(50))],
                         ids=('scalar', 'list', '0d', '1d', '3d'))
def test_scalar_aer_enu(xyz):
    """
    verify we can handle the wide variety of input data type users might use
    """
    enu = pm.ecef2enu(*xyz, 0, 90, -100)

    assert pm.enu2ecef(*enu, 0, 90, -100) == approx([0, A, 50])


def test_xarray():
    xarray = pytest.importorskip('xarray')
    xr_lla = xarray.DataArray(list(lla0))

    xyz = pm.geodetic2ecef(*xr_lla)

    assert xyz == approx(xyz0)
    assert isinstance(xyz[0], xarray.DataArray)
# %%
    xr_xyz = xarray.DataArray(list(xyz0))

    lla = pm.ecef2geodetic(*xr_xyz)

    assert lla == approx(lla0)
    assert isinstance(lla[0], float)  # xarrayness is lost, possibly expensive to keep due to isinstance()


def test_pandas():
    pandas = pytest.importorskip('pandas')
    pd_lla = pandas.Series(lla0)

    xyz = pm.geodetic2ecef(*pd_lla)

    assert xyz == approx(xyz0)
    assert isinstance(xyz[0], float)  # series degenerates to scalars by pandas itself
# %% dataframe degenerates to series
    pd_lla = pandas.DataFrame([[*lla0], [*lla0]], columns=['lat', 'lon', 'alt_m'])
    xyz = pm.geodetic2ecef(pd_lla['lat'], pd_lla['lon'], pd_lla['alt_m'])

    assert xyz[0].values == approx(xyz0[0])
    assert xyz[1].values == approx(xyz0[1])
    assert xyz[2].values == approx(xyz0[2])
    assert isinstance(xyz[0], pandas.Series)


def test_ecef():
    xyz = pm.geodetic2ecef(*lla0)

    assert xyz == approx(xyz0)
    assert pm.geodetic2ecef(*rlla0, deg=False) == approx(xyz)

    with pytest.raises(ValueError):
        pm.geodetic2ecef(-100, lla0[1], lla0[2])

    assert pm.ecef2geodetic(*xyz) == approx(lla0)
    assert pm.ecef2geodetic(*xyz, deg=False) == approx(rlla0)

    assert pm.ecef2geodetic((A - 1) / np.sqrt(2),
                            (A - 1) / np.sqrt(2), 0) == approx([0, 45, -1])


@pytest.mark.parametrize('lla, xyz', [((0, 0, -1), (A - 1, 0, 0)),
                                      ((0, 90, -1), (0, A - 1, 0)),
                                      ((0, -90, -1), (0, -A + 1, 0)),
                                      ((90, 0, -1), (0, 0, B - 1)),
                                      ((90, 15, -1), (0, 0, B - 1)),
                                      ((-90, 0, -1), (0, 0, -B + 1))
                                      ])
def test_geodetic2ecef(lla, xyz):
    assert pm.geodetic2ecef(*lla) == approx(xyz, abs=atol_dist)


@pytest.mark.parametrize('xyz, lla', [((A - 1, 0, 0), (0, 0, -1)),
                                      ((0, A - 1, 0), (0, 90, -1)),
                                      ((0, 0, B - 1), (90, 0, -1)),
                                      ((0, 0, -B + 1), (-90, 0, -1)),
                                      ((-A + 1, 0, 0), (0, 180, -1)),
                                      ])
def test_ecef2geodetic(xyz, lla):
    assert pm.ecef2geodetic(*xyz) == approx(lla)


def test_aer():
    lla2 = pm.aer2geodetic(*aer0, *lla0)
    rlla2 = pm.aer2geodetic(*raer0, *rlla0, deg=False)

    with pytest.raises(ValueError):
        pm.aer2geodetic(aer0[0], aer0[1], -1, *lla0)

    assert lla2 == approx(lla1)
    assert rlla2 == approx(rlla1)

    assert pm.geodetic2aer(*lla2, *lla0) == approx(aer0)
    assert pm.geodetic2aer(*rlla2, *rlla0, deg=False) == approx(raer0)


def test_allnan():

    anan = np.empty((10, 10))
    anan.fill(nan)
    assert np.isnan(pm.geodetic2aer(anan, anan, anan, *lla0)).all()
    assert np.isnan(pm.aer2geodetic(anan, anan, anan, *lla0)).all()


def test_somenan():
    xyz = np.stack((xyz0, (nan, nan, nan)))

    lat, lon, alt = pm.ecef2geodetic(xyz[:, 0], xyz[:, 1], xyz[:, 2])
    assert (lat[0], lon[0], alt[0]) == approx(lla0)


if __name__ == '__main__':
    pytest.main(['-x', __file__])



""" transforms involving AER: azimuth, elevation, slant range"""
from typing import Tuple
from datetime import datetime
from .ecef import ecef2enu, geodetic2ecef, ecef2geodetic, enu2uvw
from .enu import geodetic2enu, aer2enu, enu2aer
from .eci import eci2ecef, ecef2eci
from .ellipsoid import Ellipsoid

__all__ = ['aer2ecef', 'ecef2aer', 'geodetic2aer', 'aer2geodetic']


def ecef2aer(x: float, y: float, z: float,
             lat0: float, lon0: float, h0: float,
             ell: Ellipsoid = None, deg: bool = True) -> Tuple[float, float, float]:
    """
    gives azimuth, elevation and slant range from an Observer to a Point with ECEF coordinates.

    ECEF input location is with units of meters

    Parameters
    ----------

    x : float or numpy.ndarray of float
        ECEF x coordinate (meters)
    y : float or numpy.ndarray of float
        ECEF y coordinate (meters)
    z : float or numpy.ndarray of float
        ECEF z coordinate (meters)
    lat0 : float
        Observer geodetic latitude
    lon0 : float
        Observer geodetic longitude
    h0 : float
         observer altitude above geodetic ellipsoid (meters)
    ell : Ellipsoid, optional
        reference ellipsoid
    deg : bool, optional
        degrees input/output  (False: radians in/out)

    Returns
    -------
    az : float or numpy.ndarray of float
         azimuth to target
    el : float or numpy.ndarray of float
         elevation to target
    srange : float or numpy.ndarray of float
         slant range [meters]
    """
    xEast, yNorth, zUp = ecef2enu(x, y, z, lat0, lon0, h0, ell, deg=deg)

    return enu2aer(xEast, yNorth, zUp, deg=deg)


def geodetic2aer(lat: float, lon: float, h: float,
                 lat0: float, lon0: float, h0: float,
                 ell: Ellipsoid = None, deg: bool = True) -> Tuple[float, float, float]:
    """
    gives azimuth, elevation and slant range from an Observer to a Point with geodetic coordinates.


    Parameters
    ----------

    lat : float or numpy.ndarray of float
        target geodetic latitude
    lon : float or numpy.ndarray of float
        target geodetic longitude
    h : float or numpy.ndarray of float
        target altitude above geodetic ellipsoid (meters)
    lat0 : float
        Observer geodetic latitude
    lon0 : float
        Observer geodetic longitude
    h0 : float
         observer altitude above geodetic ellipsoid (meters)
    ell : Ellipsoid, optional
          reference ellipsoid
    deg : bool, optional
          degrees input/output  (False: radians in/out)

    Returns
    -------
    az : float or numpy.ndarray of float
         azimuth
    el : float or numpy.ndarray of float
         elevation
    srange : float or numpy.ndarray of float
         slant range [meters]
    """
    e, n, u = geodetic2enu(lat, lon, h, lat0, lon0, h0, ell, deg=deg)

    return enu2aer(e, n, u, deg=deg)


def aer2geodetic(az: float, el: float, srange: float,
                 lat0: float, lon0: float, h0: float,
                 ell: Ellipsoid = None,
                 deg: bool = True) -> Tuple[float, float, float]:
    """
    gives geodetic coordinates of a point with az, el, range
    from an observer at lat0, lon0, h0

    Parameters
    ----------
    az : float or numpy.ndarray of float
         azimuth to target
    el : float or numpy.ndarray of float
         elevation to target
    srange : float or numpy.ndarray of float
         slant range [meters]
    lat0 : float
           Observer geodetic latitude
    lon0 : float
           Observer geodetic longitude
    h0 : float
         observer altitude above geodetic ellipsoid (meters)
    ell : Ellipsoid, optional
          reference ellipsoid
    deg : bool, optional
          degrees input/output  (False: radians in/out)

    Returns
    -------

    In reference ellipsoid system:

    lat : float or numpy.ndarray of float
          geodetic latitude
    lon : float or numpy.ndarray of float
          geodetic longitude
    alt : float or numpy.ndarray of float
          altitude above ellipsoid  (meters)
    """
    x, y, z = aer2ecef(az, el, srange, lat0, lon0, h0, ell=ell, deg=deg)

    return ecef2geodetic(x, y, z, ell=ell, deg=deg)


def eci2aer(x: float, y: float, z: float,
            lat0: float, lon0: float, h0: float,
            t: datetime,
            useastropy: bool = True) -> Tuple[float, float, float]:
    """
    takes ECI coordinates of point and gives az, el, slant range from Observer

    Parameters
    ----------

    x : float
        ECI x-location [meters]
    y : float
        ECI y-location [meters]
    z : float
        ECI z-location [meters]
    lat0 : float
           Observer geodetic latitude
    lon0 : float
           Observer geodetic longitude
    h0 : float
         observer altitude above geodetic ellipsoid (meters)
    t : datetime.datetime
        Observation time


    Returns
    -------
    az : float
         azimuth to target
    el : float
         elevation to target
    srange : float
         slant range [meters]
    """
    xecef, yecef, zecef = eci2ecef(x, y, z, t, useastropy=useastropy)

    return ecef2aer(xecef, yecef, zecef, lat0, lon0, h0)


def aer2eci(az: float, el: float, srange: float,
            lat0: float, lon0: float, h0: float, t: datetime,
            ell=None, deg: bool = True,
            useastropy: bool = True) -> Tuple[float, float, float]:
    """
    gives ECI of a point from an observer at az, el, slant range

    Parameters
    ----------
    az : float or numpy.ndarray of float
         azimuth to target
    el : float or numpy.ndarray of float
         elevation to target
    srange : float or numpy.ndarray of float
         slant range [meters]
    lat0 : float
           Observer geodetic latitude
    lon0 : float
           Observer geodetic longitude
    h0 : float
         observer altitude above geodetic ellipsoid (meters)
    ell : Ellipsoid, optional
          reference ellipsoid
    deg : bool, optional
          degrees input/output  (False: radians in/out)
    t : datetime.datetime
        Observation time

    Returns
    -------

    Earth Centered Inertial x,y,z

    x : float or numpy.ndarray of float
        ECEF x coordinate (meters)
    y : float or numpy.ndarray of float
        ECEF y coordinate (meters)
    z : float or numpy.ndarray of float
        ECEF z coordinate (meters)
    """
    x, y, z = aer2ecef(az, el, srange, lat0, lon0, h0, ell, deg)

    return ecef2eci(x, y, z, t, useastropy=useastropy)


def aer2ecef(az: float, el: float, srange: float,
             lat0: float, lon0: float, alt0: float,
             ell: Ellipsoid = None, deg: bool = True) -> Tuple[float, float, float]:
    """
    converts target azimuth, elevation, range from observer at lat0,lon0,alt0 to ECEF coordinates.

    Parameters
    ----------
    az : float or numpy.ndarray of float
         azimuth to target
    el : float or numpy.ndarray of float
         elevation to target
    srange : float or numpy.ndarray of float
         slant range [meters]
    lat0 : float
           Observer geodetic latitude
    lon0 : float
           Observer geodetic longitude
    h0 : float
         observer altitude above geodetic ellipsoid (meters)
    ell : Ellipsoid, optional
          reference ellipsoid
    deg : bool, optional
          degrees input/output  (False: radians in/out)

    Returns
    -------

    ECEF (Earth centered, Earth fixed)  x,y,z

    x : float or numpy.ndarray of float
        ECEF x coordinate (meters)
    y : float or numpy.ndarray of float
        ECEF y coordinate (meters)
    z : float or numpy.ndarray of float
        ECEF z coordinate (meters)


    Notes
    ------
    if srange==NaN, z=NaN
    """
    # Origin of the local system in geocentric coordinates.
    x0, y0, z0 = geodetic2ecef(lat0, lon0, alt0, ell, deg=deg)
    # Convert Local Spherical AER to ENU
    e1, n1, u1 = aer2enu(az, el, srange, deg=deg)
    # Rotating ENU to ECEF
    dx, dy, dz = enu2uvw(e1, n1, u1, lat0, lon0, deg=deg)
    # Origin + offset from origin equals position in ECEF
    return x0 + dx, y0 + dy, z0 + dz



""" Line of sight intersection of space observer to ellipsoid """
from typing import Tuple
import numpy as np
from math import pi
from .aer import aer2enu
from .ecef import enu2uvw, geodetic2ecef, ecef2geodetic
from .ellipsoid import Ellipsoid

__all__ = ['lookAtSpheroid']


def lookAtSpheroid(lat0: float, lon0: float, h0: float, az: float, tilt: float,
                   ell: Ellipsoid = None, deg: bool = True) -> Tuple[float, float, float]:
    """
    Calculates line-of-sight intersection with Earth (or other ellipsoid) surface from above surface / orbit

    Parameters
    ----------

    lat0 : float
           observer geodetic latitude
    lon0 : float
           observer geodetic longitude
    h0 : float
        observer altitude (meters)  Must be non-negative since this function doesn't consider terrain
    az : float or numpy.ndarray of float
        azimuth angle of line-of-sight, clockwise from North
    tilt : float or numpy.ndarray of float
        tilt angle of line-of-sight with respect to local vertical (nadir = 0)
    ell : Ellipsoid, optional
          reference ellipsoid
    deg : bool, optional
          degrees input/output  (False: radians in/out)

    Results
    -------

    lat0 : float or numpy.ndarray of float
           geodetic latitude where the line-of-sight intersects with the Earth ellipsoid
    lon0 : float or numpy.ndarray of float
           geodetic longitude where the line-of-sight intersects with the Earth ellipsoid
    d : float or numpy.ndarray of float
        slant range (meters) from starting point to intersect point

    Values will be NaN if the line of sight does not intersect.

    Algorithm based on https://medium.com/@stephenhartzell/satellite-line-of-sight-intersection-with-earth-d786b4a6a9b6 Stephen Hartzell
    """

    if (np.asarray(h0) < 0).any():
        raise ValueError('Intersection calculation requires altitude  [0, Infinity)')

    if ell is None:
        ell = Ellipsoid()

    tilt = np.asarray(tilt)

    a = ell.semimajor_axis
    b = ell.semimajor_axis
    c = ell.semiminor_axis

    el = tilt - 90. if deg else tilt - pi / 2

    e, n, u = aer2enu(az, el, srange=1., deg=deg)  # fixed 1 km slant range
    u, v, w = enu2uvw(e, n, u, lat0, lon0, deg=deg)
    x, y, z = geodetic2ecef(lat0, lon0, h0, deg=deg)

    value = -a**2 * b**2 * w * z - a**2 * c**2 * v * y - b**2 * c**2 * u * x
    radical = (a**2 * b**2 * w**2 + a**2 * c**2 * v**2 - a**2 * v**2 * z**2 + 2 * a**2 * v * w * y * z -
               a**2 * w**2 * y**2 + b**2 * c**2 * u**2 - b**2 * u**2 * z**2 + 2 * b**2 * u * w * x * z -
               b**2 * w**2 * x**2 - c**2 * u**2 * y**2 + 2 * c**2 * u * v * x * y - c**2 * v**2 * x**2)

    magnitude = a**2 * b**2 * w**2 + a**2 * c**2 * v**2 + b**2 * c**2 * u**2

# %%   Return nan if radical < 0 or d < 0 because LOS vector does not point towards Earth
    with np.errstate(invalid='ignore'):
        d = np.where(radical > 0,
                     (value - a * b * c * np.sqrt(radical)) / magnitude,
                     np.nan)
        d[d < 0] = np.nan
# %% cartesian to ellipsodal
    lat, lon, _ = ecef2geodetic(x + d * u, y + d * v, z + d * w, deg=deg)

    return lat, lon, d



#!/usr/bin/env python
import pytest
from pytest import approx
import numpy as np
import pymap3d as pm
import pymap3d.sidereal as pmd
import pymap3d.haversine as pmh

lon = -148
t0 = '2014-04-06T08:00:00'
lla0 = (42, -82, 200)
sra = 2.90658
ha = 45.482789587392013
eci0 = (-3.977913815668146e6, -2.582332196263046e6, 4.250818828152067e6)


@pytest.mark.parametrize('time', [t0, [t0]], ids=('scalar', 'list'))
def test_sidereal(time):
    pytest.importorskip('astropy')
    # http://www.jgiesen.de/astro/astroJS/siderealClock/
    assert pmd.datetime2sidereal(time, np.radians(lon), False) == approx(sra, rel=1e-5)


@pytest.mark.parametrize('time', [t0, [t0]], ids=('scalar', 'list'))
def test_sidereal_vallado(time):
    assert pmd.datetime2sidereal(time, np.radians(lon), True) == approx(sra, rel=1e-5)


def test_anglesep():
    pytest.importorskip('astropy')

    assert pmh.anglesep(35, 23, 84, 20) == approx(ha)


def test_anglesep_meeus():
    # %% compare with astropy
    assert pmh.anglesep_meeus(35, 23, 84, 20) == approx(ha)


@pytest.mark.parametrize('useastropy', [True, False])
def test_eci_geodetic(useastropy):
    t = '2013-01-15T12:00:05'
    lla = pm.eci2geodetic(*eci0, t, useastropy=useastropy)
    assert lla == approx(lla0, rel=0.2)


@pytest.mark.parametrize('useastropy', [True, False])
def test_eci_aer(useastropy):
    t = '2013-01-15T12:00:05'

    aer1 = pm.eci2aer(*eci0, 42, -100, 0, t, useastropy=useastropy)
    assert aer1 == approx([83.73050, -6.614478, 1.473510e6], rel=0.001)

    assert pm.aer2eci(*aer1, 42, -100, 0, t, useastropy=useastropy) == approx(eci0, rel=0.001)

    with pytest.raises(ValueError):
        pm.aer2eci(aer1[0], aer1[1], -1, 42, -100, 0, t, useastropy=useastropy)


if __name__ == '__main__':
    pytest.main(['-xrsv', __file__])



"""
for strictly pure Python (no Numpy) only modules
Python >= 3.5
"""
import pytest
from pytest import approx
from math import radians

import pymap3d.math as pm

lla0 = (42, -82, 200)
rlla0 = (radians(lla0[0]), radians(lla0[1]), lla0[2])
aer0 = (33, 70, 1000)
raer0 = (radians(aer0[0]), radians(aer0[1]), aer0[2])
enu0 = (186.277521, 286.842228, 939.692621)
ned0 = (enu0[1], enu0[0], -enu0[2])
vx, vy, vz = (5, 3, 2)
ve, vn, vu = (5.368859646588048, 3.008520763668120, -0.352347711524077)


def test_aer_enu():
    xyz = pm.aer2ecef(*aer0, *lla0)

    enu = pm.aer2enu(*aer0)

    assert enu == approx(enu0)
    assert pm.aer2enu(*raer0, deg=False) == approx(enu0)

    with pytest.raises(ValueError):
        pm.aer2enu(aer0[0], aer0[1], -1)

    assert pm.enu2ecef(*enu, *lla0) == approx(xyz)
    assert pm.enu2ecef(*enu, *rlla0, deg=False) == approx(xyz)

    assert pm.ecef2enu(*xyz, *lla0) == approx(enu)
    assert pm.ecef2enu(*xyz, *rlla0, deg=False) == approx(enu)


def test_ned():
    xyz = pm.aer2ecef(*aer0, *lla0)
    enu = pm.aer2enu(*aer0)
    ned = (enu[1], enu[0], -enu[2])
    lla = pm.aer2geodetic(*aer0, *lla0)

    assert pm.aer2ned(*aer0) == approx(ned0)

    with pytest.raises(ValueError):
        pm.aer2ned(aer0[0], aer0[1], -1)

    assert pm.enu2aer(*enu) == approx(aer0)
    assert pm.enu2aer(*enu, deg=False) == approx(raer0)

    assert pm.ned2aer(*ned) == approx(aer0)

    assert pm.ecef2ned(*xyz, *lla0) == approx(ned)

    assert pm.ned2ecef(*ned, *lla0) == approx(xyz)
# %%
    assert pm.ecef2enuv(vx, vy, vz, *lla0[:2]) == approx((ve, vn, vu))

    assert pm.ecef2nedv(vx, vy, vz, *lla0[:2]) == approx((vn, ve, -vu))

# %%
    enu3 = pm.geodetic2enu(*lla, *lla0)
    ned3 = (enu3[1], enu3[0], -enu3[2])

    assert pm.geodetic2ned(*lla, *lla0) == approx(ned3)

    assert pm.enu2geodetic(*enu3, *lla0) == approx(lla)

    assert pm.ned2geodetic(*ned3, *lla0) == approx(lla)


if __name__ == '__main__':
    pytest.main([__file__])



#!/usr/bin/env python
# Copyright (c) 2014-2018 Michael Hirsch, Ph.D.

"""
converts right ascension, declination to azimuth, elevation and vice versa.
Normally do this via AstroPy.
These functions are fallbacks for those wihtout AstroPy.

Michael Hirsch implementation of algorithms from D. Vallado
"""
from datetime import datetime
from numpy import sin, cos, degrees, radians, arcsin, arctan2, atleast_1d
from typing import Tuple
from .sidereal import datetime2sidereal

__all__ = ['azel2radec', 'radec2azel']


def azel2radec(az_deg: float, el_deg: float,
               lat_deg: float, lon_deg: float,
               time: datetime) -> Tuple[float, float]:
    """
    converts azimuth, elevation to right ascension, declination

    Parameters
    ----------

    az_deg : float or numpy.ndarray of float
        azimuth (clockwise) to point [degrees]

    el_deg : float or numpy.ndarray of float
        elevation above horizon to point [degrees]

    lat_deg : float
        observer WGS84 latitude [degrees]

    lon_deg : float
        observer WGS84 longitude [degrees]

    time : datetime.datetime
        time of observation

    Results
    -------

    ra_deg : float or numpy.ndarray of float
        right ascension to target [degrees]

    dec_deg : float or numpy.ndarray of float
        declination of target [degrees]

    from D.Vallado Fundamentals of Astrodynamics and Applications
    p.258-259
    """
    az = atleast_1d(az_deg)
    el = atleast_1d(el_deg)
    lat = atleast_1d(lat_deg)
    lon = atleast_1d(lon_deg)

    if az.shape != el.shape:
        raise ValueError('az and el must be same shape ndarray')
    if not(lat.size == 1 and lon.size == 1):
        raise ValueError('need one observer and one or more  (az,el).')
    if ((lat < -90) | (lat > 90)).any():
        raise ValueError('-90 <= lat <= 90')

    az = radians(az)
    el = radians(el)
    lat = radians(lat)
    lon = radians(lon)
# %% Vallado "algorithm 28" p 268
    dec = arcsin(sin(el) * sin(lat) + cos(el) * cos(lat) * cos(az))

    lha = arctan2(-(sin(az) * cos(el)) / cos(dec),
                  (sin(el) - sin(lat) * sin(dec)) / (cos(dec) * cos(lat)))

    lst = datetime2sidereal(time, lon)  # lon, ra in RADIANS

    """ by definition right ascension [0, 360) degrees """
    return degrees(lst - lha) % 360, degrees(dec)


def radec2azel(ra_deg: float, dec_deg: float,
               lat_deg: float, lon_deg: float,
               time: datetime) -> Tuple[float, float]:
    """
    converts right ascension, declination to azimuth, elevation

    Parameters
    ----------

    ra_deg : float or numpy.ndarray of float
        right ascension to target [degrees]

    dec_deg : float or numpy.ndarray of float
        declination to target [degrees]

    lat_deg : float
        observer WGS84 latitude [degrees]

    lon_deg : float
        observer WGS84 longitude [degrees]

    time : datetime.datetime
        time of observation

    Results
    -------

    az_deg : float or numpy.ndarray of float
        azimuth clockwise from north to point [degrees]

    el_deg : float or numpy.ndarray of float
        elevation above horizon to point [degrees]


    from D. Vallado "Fundamentals of Astrodynamics and Applications "
       4th Edition Ch. 4.4 pg. 266-268
    """
    ra = atleast_1d(ra_deg)
    dec = atleast_1d(dec_deg)
    lat = atleast_1d(lat_deg)
    lon = atleast_1d(lon_deg)

    if ra.shape != dec.shape:
        raise ValueError('az and el must be same shape ndarray')
    if not(lat.size == 1 and lon.size == 1):
        raise ValueError('need one observer and one or more  (az,el).')
    if ((lat < -90) | (lat > 90)).any():
        raise ValueError('-90 <= lat <= 90')

    ra = radians(ra)
    dec = radians(dec)
    lat = radians(lat)
    lon = radians(lon)

    lst = datetime2sidereal(time, lon)  # RADIANS
# %% Eq. 4-11 p. 267 LOCAL HOUR ANGLE
    lha = lst - ra
# %% #Eq. 4-12 p. 267
    el = arcsin(sin(lat) * sin(dec) + cos(lat) * cos(dec) * cos(lha))
# %% combine Eq. 4-13 and 4-14 p. 268
    az = arctan2(-sin(lha) * cos(dec) / cos(el),
                 (sin(dec) - sin(el) * sin(lat)) / (cos(el) * cos(lat)))

    return degrees(az) % 360.0, degrees(el)



# Copyright (c) 2014-2018 Michael Hirsch, Ph.D.
""" convert strings to datetime """
from datetime import datetime
from dateutil.parser import parse
import numpy as np

__all__ = ['str2dt']


def str2dt(time: datetime) -> np.ndarray:
    """
    Converts times in string or list of strings to datetime(s)

    Parameters
    ----------

    time : str or datetime.datetime or numpy.datetime64

    Results
    -------

    t : datetime.datetime

    """
    if isinstance(time, datetime):
        return time
    elif isinstance(time, str):
        return parse(time)
    elif isinstance(time, np.datetime64):
        return time.astype(datetime)
    else:  # some sort of iterable
        try:
            if isinstance(time[0], datetime):
                return time
            elif isinstance(time[0], np.datetime64):
                return time.astype(datetime)
            elif isinstance(time[0], str):
                return [parse(t) for t in time]
        except (IndexError, TypeError):
            pass

        # last resort--assume pandas/xarray

        return time.values.astype('datetime64[us]').astype(datetime)



#!/usr/bin/env python
"""
vreckon and vdist are iterative algorithms.
How much does PyPy help over Cpython?

Hmm, PyPy is slower than Cpython..

$ pypy3 tests/benchmark_vincenty.py 10000
2.1160879135131836
0.06056046485900879

$ python tests/benchmark_vincenty.py 10000
0.3325080871582031
0.02107095718383789
"""
from time import time
from pymap3d.vincenty import vreckon, vdist
import numpy as np
from argparse import ArgumentParser

ll0 = (42., 82.)


def bench_vreckon(N: int) -> float:

    sr = np.random.random(N)
    az = np.random.random(N)

    tic = time()
    a, b, c = vreckon(*ll0, sr, az)

    return time() - tic


def bench_vdist(N: int) -> float:
    lat = np.random.random(N)
    lon = np.random.random(N)

    tic = time()
    asr, aaz, aa21 = vdist(*ll0, lat, lon)

    return time() - tic


if __name__ == '__main__':
    p = ArgumentParser()
    p.add_argument('N', type=int)
    p = p.parse_args()

    print(bench_vreckon(p.N))
    print(bench_vdist(p.N))



#!/usr/bin/env python
from pymap3d.vincenty import vreckon
from argparse import ArgumentParser


def main():
    p = ArgumentParser(description='Given starting latitude, longitude: find final lat,lon for distance and azimuth')
    p.add_argument('lat', help='latitude WGS-84 [degrees]', type=float)
    p.add_argument('lon', help='longitude WGS-84 [degrees]', type=float)
    p.add_argument('range', help='range from start point [meters]', type=float)
    p.add_argument('azimuth', help='clockwise from north: azimuth to start [degrees]', type=float)
    P = p.parse_args()

    lat2, lon2, a21 = vreckon(P.lat, P.lon, P.range, P.azimuth)

    print('{:.4f} {:.4f}'.format(lat2, lon2))
    print('{:.1f}'.format(a21))


if __name__ == '__main__':  # pragma: no cover
    main()



""" Transforms involving ECEF: earth-centered, earth-fixed frame """
from math import radians, sin, cos, tan, hypot, degrees, sqrt, pi
from math import atan as arctan, atan2 as arctan2
from ..ellipsoid import Ellipsoid

tau = 2 * pi

__all__ = ['geodetic2ecef', 'ecef2geodetic', 'ecef2enuv', 'ecef2enu', 'enu2uvw', 'uvw2enu', 'enu2ecef']


# def geodetic2ecef(lat: float, lon: float, alt: float, ell: Ellipsoid = None, deg: bool = True):
def geodetic2ecef(lat, lon, alt, ell=None, deg=True):
    """
    point transformation from Geodetic of specified ellipsoid (default WGS-84) to ECEF

    Parameters
    ----------

    lat : float
           target geodetic latitude
    lon : float
           target geodetic longitude
    h : float
         target altitude above geodetic ellipsoid (meters)
    ell : Ellipsoid, optional
          reference ellipsoid
    deg : bool, optional
          degrees input/output  (False: radians in/out)


    Returns
    -------

    ECEF (Earth centered, Earth fixed)  x,y,z

    x : float
        target x ECEF coordinate (meters)
    y : float
        target y ECEF coordinate (meters)
    z : float
        target z ECEF coordinate (meters)
    """
    if ell is None:
        ell = Ellipsoid()

    if deg:
        lat = radians(lat)
        lon = radians(lon)

    if (lat < -pi / 2) | (lat > pi / 2):
        raise ValueError('-90 <= lat <= 90')

    # radius of curvature of the prime vertical section
    N = ell.semimajor_axis**2 / sqrt(ell.semimajor_axis**2 * cos(lat)**2 + ell.semiminor_axis**2 * sin(lat)**2)
    # Compute cartesian (geocentric) coordinates given  (curvilinear) geodetic
    # coordinates.
    x = (N + alt) * cos(lat) * cos(lon)
    y = (N + alt) * cos(lat) * sin(lon)
    z = (N * (ell.semiminor_axis / ell.semimajor_axis)**2 + alt) * sin(lat)

    return x, y, z


# def ecef2geodetic(x: float, y: float, z: float, ell: Ellipsoid = None, deg: bool = True):
def ecef2geodetic(x, y, z, ell=None, deg=True):
    """
    convert ECEF (meters) to geodetic coordinates

    Parameters
    ----------
    x : float
        target x ECEF coordinate (meters)
    y : float
        target y ECEF coordinate (meters)
    z : float
        target z ECEF coordinate (meters)
    ell : Ellipsoid, optional
          reference ellipsoid
    deg : bool, optional
          degrees input/output  (False: radians in/out)

    Returns
    -------
    lat : float
           target geodetic latitude
    lon : float
           target geodetic longitude
    h : float
         target altitude above geodetic ellipsoid (meters)

    based on:
    You, Rey-Jer. (2000). Transformation of Cartesian to Geodetic Coordinates without Iterations.
    Journal of Surveying Engineering. doi: 10.1061/(ASCE)0733-9453
    """
    if ell is None:
        ell = Ellipsoid()

    r = sqrt(x**2 + y**2 + z**2)

    E = sqrt(ell.semimajor_axis**2 - ell.semiminor_axis**2)

    # eqn. 4a
    u = sqrt(0.5 * (r**2 - E**2) + 0.5 * sqrt((r**2 - E**2)**2 + 4 * E**2 * z**2))

    Q = hypot(x, y)

    huE = hypot(u, E)

    # eqn. 4b
    Beta = arctan(huE / u * z / hypot(x, y))

    # eqn. 13
    eps = ((ell.semiminor_axis * u - ell.semimajor_axis * huE + E**2) * sin(Beta)) / (ell.semimajor_axis * huE * 1 / cos(Beta) - E**2 * cos(Beta))

    Beta += eps
# %% final output
    lat = arctan(ell.semimajor_axis / ell.semiminor_axis * tan(Beta))

    lon = arctan2(y, x)

    # eqn. 7
    alt = hypot(z - ell.semiminor_axis * sin(Beta),
                Q - ell.semimajor_axis * cos(Beta))

    # inside ellipsoid?
    inside = x**2 / ell.semimajor_axis**2 + y**2 / ell.semimajor_axis**2 + z**2 / ell.semiminor_axis**2 < 1
    if inside:
        alt = -alt

    if deg:
        lat = degrees(lat)
        lon = degrees(lon)

    return lat, lon, alt


# def ecef2enuv(u: float, v: float, w: float, lat0: float, lon0: float, deg: bool = True):
def ecef2enuv(u, v, w, lat0, lon0, deg=True):
    """
    VECTOR from observer to target  ECEF => ENU

    Parameters
    ----------
    u : float
        target x ECEF coordinate (meters)
    v : float
        target y ECEF coordinate (meters)
    w : float
        target z ECEF coordinate (meters)
    lat0 : float
           Observer geodetic latitude
    lon0 : float
           Observer geodetic longitude
    h0 : float
         observer altitude above geodetic ellipsoid (meters)
    deg : bool, optional
          degrees input/output  (False: radians in/out)

    Returns
    -------
    uEast : float
        target east ENU coordinate (meters)
    vNorth : float
        target north ENU coordinate (meters)
    wUp : float
        target up ENU coordinate (meters)

    """
    if deg:
        lat0 = radians(lat0)
        lon0 = radians(lon0)

    t = cos(lon0) * u + sin(lon0) * v
    uEast = -sin(lon0) * u + cos(lon0) * v
    wUp = cos(lat0) * t + sin(lat0) * w
    vNorth = -sin(lat0) * t + cos(lat0) * w

    return uEast, vNorth, wUp


# def ecef2enu(x: float, y: float, z: float, lat0: float, lon0: float, h0: float, ell: Ellipsoid = None, deg: bool = True):
def ecef2enu(x, y, z, lat0, lon0, h0, ell=None, deg=True):
    """
    from observer to target, ECEF => ENU

    Parameters
    ----------
    x : float
        target x ECEF coordinate (meters)
    y : float
        target y ECEF coordinate (meters)
    z : float
        target z ECEF coordinate (meters)
    lat0 : float
           Observer geodetic latitude
    lon0 : float
           Observer geodetic longitude
    h0 : float
         observer altitude above geodetic ellipsoid (meters)
    ell : Ellipsoid, optional
          reference ellipsoid
    deg : bool, optional
          degrees input/output  (False: radians in/out)

    Returns
    -------
    East : float
        target east ENU coordinate (meters)
    North : float
        target north ENU coordinate (meters)
    Up : float
        target up ENU coordinate (meters)

    """
    x0, y0, z0 = geodetic2ecef(lat0, lon0, h0, ell, deg=deg)

    return uvw2enu(x - x0, y - y0, z - z0, lat0, lon0, deg=deg)


# def enu2uvw(east: float, north: float, up: float, lat0: float, lon0: float, deg: bool = True):
def enu2uvw(east, north, up, lat0, lon0, deg=True):
    """
    Parameters
    ----------

    e1 : float
        target east ENU coordinate (meters)
    n1 : float
        target north ENU coordinate (meters)
    u1 : float
        target up ENU coordinate (meters)

    Results
    -------

    u : float
    v : float
    w : float
    """

    if deg:
        lat0 = radians(lat0)
        lon0 = radians(lon0)

    t = cos(lat0) * up - sin(lat0) * north
    w = sin(lat0) * up + cos(lat0) * north

    u = cos(lon0) * t - sin(lon0) * east
    v = sin(lon0) * t + cos(lon0) * east

    return u, v, w


# def uvw2enu(u: float, v: float, w: float, lat0: float, lon0: float, deg: bool = True):
def uvw2enu(u, v, w, lat0, lon0, deg=True):
    """
    Parameters
    ----------

    u : float
    v : float
    w : float


    Results
    -------

    East : float
        target east ENU coordinate (meters)
    North : float
        target north ENU coordinate (meters)
    Up : float
        target up ENU coordinate (meters)
    """
    if deg:
        lat0 = radians(lat0)
        lon0 = radians(lon0)

    t = cos(lon0) * u + sin(lon0) * v
    East = -sin(lon0) * u + cos(lon0) * v
    Up = cos(lat0) * t + sin(lat0) * w
    North = -sin(lat0) * t + cos(lat0) * w

    return East, North, Up


# def enu2ecef(e1: float, n1: float, u1: float, lat0: float, lon0: float, h0: float, ell: Ellipsoid = None, deg: bool = True):
def enu2ecef(e1, n1, u1, lat0, lon0, h0, ell=None, deg=True):
    """
    ENU to ECEF

    Parameters
    ----------

    e1 : float
        target east ENU coordinate (meters)
    n1 : float
        target north ENU coordinate (meters)
    u1 : float
        target up ENU coordinate (meters)
    lat0 : float
           Observer geodetic latitude
    lon0 : float
           Observer geodetic longitude
    h0 : float
         observer altitude above geodetic ellipsoid (meters)
    ell : Ellipsoid, optional
          reference ellipsoid
    deg : bool, optional
          degrees input/output  (False: radians in/out)


    Results
    -------
    x : float
        target x ECEF coordinate (meters)
    y : float
        target y ECEF coordinate (meters)
    z : float
        target z ECEF coordinate (meters)
    """
    x0, y0, z0 = geodetic2ecef(lat0, lon0, h0, ell, deg=deg)
    dx, dy, dz = enu2uvw(e1, n1, u1, lat0, lon0, deg=deg)

    return x0 + dx, y0 + dy, z0 + dz



from .ecef import Ellipsoid, geodetic2ecef, ecef2geodetic, ecef2enuv, ecef2enu, enu2uvw, uvw2enu, enu2ecef  # noqa: F401
from .enu import enu2aer, aer2enu, enu2geodetic, geodetic2enu  # noqa: F401
from .aer import aer2ecef, ecef2aer, geodetic2aer, aer2geodetic  # noqa: F401
from .ned import aer2ned, ned2aer, ned2geodetic, ned2ecef, ecef2ned, geodetic2ned, ecef2nedv  # noqa: F401



#!/usr/bin/env python
"""
Example of using Google Maps queries and PyMap3D

https://developers.google.com/places/web-service/search

This requires a Google Cloud key, and costs a couple US cents per query.

TODO: Would like to instead query a larger region, would OSM be an option?
"""
from pathlib import Path
from argparse import ArgumentParser
import requests
import functools
import pandas

from pymap3d.vincenty import vdist


URL = 'https://maps.googleapis.com/maps/api/place/nearbysearch/json?'


@functools.lru_cache()
def get_place_coords(place_type: str,
                     latitude: float, longitude: float,
                     search_radius_km: int,
                     keyfn: Path) -> pandas.DataFrame:
    """
    Get places using Google Maps Places API
    Requires you to have a Google Cloud account with API key.
    """

    keyfn = Path(keyfn).expanduser()
    key = keyfn.read_text()

    stub = URL + 'location={},{}'.format(latitude, longitude)

    stub += '&radius={}'.format(search_radius_km * 1000)

    stub += '&types={}'.format(place_type)

    stub += '&key={}'.format(key)

    r = requests.get(stub)
    r.raise_for_status()

    place_json = r.json()['results']

    places = pandas.DataFrame(index=[p['name'] for p in place_json],
                              columns=['latitude', 'longitude', 'distance_km', 'vicinity'])
    places['latitude'] = [p['geometry']['location']['lat'] for p in place_json]
    places['longitude'] = [p['geometry']['location']['lng'] for p in place_json]
    places['vicinity'] = [p['vicinity'] for p in place_json]

    return places


if __name__ == '__main__':
    p = ArgumentParser()
    p.add_argument('place_type', help='Place type to search: https://developers.google.com/places/supported_types')
    p.add_argument('searchloc', help='initial latituude, longitude to search from', nargs=2, type=float)
    p.add_argument('radius', help='search radius (kilometers)', type=int)
    p.add_argument('refloc', help='reference location (lat, lon)', nargs=2, type=float)
    p.add_argument('-k', '--keyfn', help='Google Places API key file', default='~/googlemaps.key')
    a = p.parse_args()

    place_coords = get_place_coords(a.place_type, *a.searchloc, a.radius, a.keyfn)

    place_coords['distance_km'] = vdist(place_coords['latitude'], place_coords['longitude'], *a.refloc)[0] / 1e3



#!/usr/bin/env python
# Copyright (c) 2014-2018 Michael Hirsch, Ph.D.
"""
PyMap3D provides coordinate transforms and geodesy functions with a similar API
to the Matlab Mapping Toolbox, but was of course independently derived.

For all functions, the default units are:

distance : float
    METERS
angles : float
    DEGREES
time : datetime.datetime
    UTC time of observation

These functions may be used with any planetary body, provided the appropriate
reference ellipsoid is defined. The default ellipsoid is WGS-84

deg : bool = True means degrees. False = radians.

Most functions accept NumPy arrays of any shape, as well as compatible data types
including AstroPy, Pandas and Xarray that have Numpy-like data properties.
For clarity, we omit all these types in the docs, and just specify the scalar type.

Other languages
---------------

Companion packages exist for:

* Matlab / GNU Octave: [Matmap3D](https://github.com/scivision/matmap3d)
* Fortran: [Maptran3D](https://github.com/scivision/maptran3d)
"""
import sys

if sys.version_info >= (3, 5):
    try:
        import numpy
        import dateutil  # noqa: F401
    except ImportError:
        numpy = None
else:
    numpy = None

if numpy is not None:
    from .timeconv import str2dt  # noqa: F401
    from .azelradec import radec2azel, azel2radec  # noqa: F401
    from .eci import eci2ecef, ecef2eci  # noqa: F401
    from .sidereal import datetime2sidereal  # noqa: F401
    from .ecef import geodetic2ecef, ecef2geodetic, eci2geodetic, ecef2enuv, enu2ecef, ecef2enu, enu2uvw, uvw2enu  # noqa: F401
    from .ellipsoid import Ellipsoid  # noqa: F401
    from .ned import ned2ecef, ned2geodetic, geodetic2ned, ecef2nedv, ned2aer, aer2ned, ecef2ned  # noqa: F401
    from .enu import enu2geodetic, geodetic2enu, aer2enu, enu2aer  # noqa: F401
    from .aer import ecef2aer, aer2ecef, geodetic2aer, aer2geodetic, eci2aer, aer2eci  # noqa: F401
    from .los import lookAtSpheroid  # noqa: F401
    from .lox import isometric, meridian_dist, loxodrome_inverse  # noqa: F401
else:  # pure Python only
    from .math import *  # type: ignore  # noqa: F401, F403





class Ellipsoid:
    """
    generate reference ellipsoid parameters

    https://en.wikibooks.org/wiki/PROJ.4#Spheroid

    https://nssdc.gsfc.nasa.gov/planetary/factsheet/index.html

    as everywhere else in this program, distance units are METERS
    """

    # def __init__(self, model: str = 'wgs84'):
    def __init__(self, model='wgs84'):
        """
        feel free to suggest additional ellipsoids

        Parameters
        ----------
        model : str
                name of ellipsoid
        """
        if model == 'wgs84':
            """https://en.wikipedia.org/wiki/World_Geodetic_System#WGS84"""
            self.semimajor_axis = 6378137.
            self.flattening = 1 / 298.2572235630
            self.semiminor_axis = self.semimajor_axis * (1 - self.flattening)
        elif model == 'wgs72':
            self.semimajor_axis = 6378135.
            self.flattening = 1 / 298.26
            self.semiminor_axis = self.semimajor_axis * (1 - self.flattening)
        elif model == 'grs80':
            """https://en.wikipedia.org/wiki/GRS_80"""
            self.semimajor_axis = 6378137.
            self.flattening = 1 / 298.257222100882711243
            self.semiminor_axis = self.semimajor_axis * (1 - self.flattening)
        elif model == 'clarke1866':
            self.semimajor_axis = 6378206.4
            self.semiminor_axis = 6356583.8
            self.flattening = -(self.semiminor_axis / self.semimajor_axis - 1)
        elif model == 'mars':
            """
            https://tharsis.gsfc.nasa.gov/geodesy.html
            """
            self.semimajor_axis = 3396900.
            self.semiminor_axis = 3376097.80585952
            self.flattening = 1 / 163.295274386012
        elif model == 'moon':
            self.semimajor_axis = 1738000.
            self.semiminor_axis = self.semimajor_axis
            self.flattening = 0.
        elif model == 'venus':
            self.semimajor_axis = 6051000.
            self.semiminor_axis = self.semimajor_axis
            self.flattening = 0.
        elif model == 'jupiter':
            self.semimajor_axis = 71492000.
            self.flattening = 1 / 15.415446277169725
            self.flattening = -(self.semiminor_axis / self.semimajor_axis - 1)
        elif model == 'io':
            """
            https://doi.org/10.1006/icar.1998.5987
            """
            self.semimajor_axis = 1829.7
            self.flattening = 1 / 131.633093525179
            self.semiminor_axis = self.semimajor_axis * (1 - self.flattening)
        elif model == 'pluto':
            self.semimajor_axis = 1187000.
            self.semiminor_axis = self.semimajor_axis
            self.flattening = 0.
        else:
            raise NotImplementedError('{} model not implemented, let us know and we will add it (or make a pull request)'.format(model))



""" transforms involving ENU East North Up """
from typing import Tuple
from numpy import radians, sin, cos, hypot, arctan2, degrees
import numpy as np
try:
    from math import tau  # Python >= 3.6
except ImportError:
    tau = 2 * np.pi

from .ecef import geodetic2ecef, ecef2geodetic, enu2ecef, uvw2enu
from .ellipsoid import Ellipsoid


__all__ = ['enu2aer', 'aer2enu', 'enu2geodetic', 'geodetic2enu']


def enu2aer(e: float, n: float, u: float, deg: bool = True) -> Tuple[float, float, float]:
    """
    ENU to Azimuth, Elevation, Range

    Parameters
    ----------

    e : float or np.ndarray of float
        ENU East coordinate (meters)
    n : float or np.ndarray of float
        ENU North coordinate (meters)
    u : float or np.ndarray of float
        ENU Up coordinate (meters)
    deg : bool, optional
        degrees input/output  (False: radians in/out)

    Results
    -------

    azimuth : float or np.ndarray of float
        azimuth to rarget
    elevation : float or np.ndarray of float
        elevation to target
    srange : float or np.ndarray of float
        slant range [meters]
    """
    # 1 millimeter precision for singularity

    e = np.asarray(e)
    n = np.asarray(n)
    u = np.asarray(u)

    with np.errstate(invalid='ignore'):
        e[abs(e) < 1e-3] = 0.
        n[abs(n) < 1e-3] = 0.
        u[abs(u) < 1e-3] = 0.

        r = hypot(e, n)
        slantRange = hypot(r, u)
        elev = arctan2(u, r)
        az = arctan2(e, n) % tau

    if deg:
        az = degrees(az)
        elev = degrees(elev)

    return az, elev, slantRange


def aer2enu(az: float, el: float, srange: float,
            deg: bool = True) -> Tuple[float, float, float]:
    """
    Azimuth, Elevation, Slant range to target to East, north, Up

    Parameters
    ----------
    azimuth : float or np.ndarray of float
            azimuth clockwise from north (degrees)
    elevation : float or np.ndarray of float
        elevation angle above horizon, neglecting aberattions (degrees)
    srange : float or np.ndarray of float
        slant range [meters]
    deg : bool, optional
        degrees input/output  (False: radians in/out)

    Returns
    --------
    e : float or np.ndarray of float
        East ENU coordinate (meters)
    n : float or np.ndarray of float
        North ENU coordinate (meters)
    u : float or np.ndarray of float
        Up ENU coordinate (meters)
    """
    if deg:
        el = radians(el)
        az = radians(az)

    with np.errstate(invalid='ignore'):
        if (np.asarray(srange) < 0).any():
            raise ValueError('Slant range  [0, Infinity)')

    r = srange * cos(el)

    return r * sin(az), r * cos(az), srange * sin(el)


def enu2geodetic(e: float, n: float, u: float,
                 lat0: float, lon0: float, h0: float,
                 ell: Ellipsoid = None, deg: bool = True) -> Tuple[float, float, float]:
    """
    East, North, Up to target to geodetic coordinates

    Parameters
    ----------
    e : float or np.ndarray of float
        East ENU coordinate (meters)
    n : float or np.ndarray of float
        North ENU coordinate (meters)
    u : float or np.ndarray of float
        Up ENU coordinate (meters)
    lat0 : float
           Observer geodetic latitude
    lon0 : float
           Observer geodetic longitude
    h0 : float
         observer altitude above geodetic ellipsoid (meters)
    ell : Ellipsoid, optional
          reference ellipsoid
    deg : bool, optional
          degrees input/output  (False: radians in/out)


    Results
    -------
    lat : float or np.ndarray of float
          geodetic latitude
    lon : float or np.ndarray of float
          geodetic longitude
    alt : float or np.ndarray of float
          altitude above ellipsoid  (meters)
    """

    x, y, z = enu2ecef(e, n, u, lat0, lon0, h0, ell, deg=deg)

    return ecef2geodetic(x, y, z, ell, deg=deg)


def geodetic2enu(lat: float, lon: float, h: float,
                 lat0: float, lon0: float, h0: float,
                 ell: Ellipsoid = None, deg: bool = True) -> Tuple[float, float, float]:
    """
    Parameters
    ----------
    lat : float or np.ndarray of float
          target geodetic latitude
    lon : float or np.ndarray of float
          target geodetic longitude
    h : float or np.ndarray of float
          target altitude above ellipsoid  (meters)
    lat0 : float
           Observer geodetic latitude
    lon0 : float
           Observer geodetic longitude
    h0 : float
         observer altitude above geodetic ellipsoid (meters)
    ell : Ellipsoid, optional
          reference ellipsoid
    deg : bool, optional
          degrees input/output  (False: radians in/out)


    Results
    -------
    e : float or np.ndarray of float
        East ENU
    n : float or np.ndarray of float
        North ENU
    u : float or np.ndarray of float
        Up ENU
    """
    x1, y1, z1 = geodetic2ecef(lat, lon, h, ell, deg=deg)
    x2, y2, z2 = geodetic2ecef(lat0, lon0, h0, ell, deg=deg)

    return uvw2enu(x1 - x2, y1 - y2, z1 - z2,
                   lat0, lon0, deg=deg)
